# (C) Copyright 2020- ECMWF.  #
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from abc import abstractmethod

LOG = logging.getLogger(__name__)

PRIVATE_ATTRIBUTES = {"observer": lambda: None}


class MetaBase(type):
    def __call__(cls, *args, **kwargs):
        obj = cls.__new__(cls, *args, **kwargs)
        args, kwargs = cls.patch(obj, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        return obj

    def patch(cls, obj, *args, **kwargs):
        private_attributes = {}
        private_attributes.update(PRIVATE_ATTRIBUTES)
        private_attributes.update(kwargs.pop("_PRIVATE_ATTRIBUTES", {}))

        for k, v in private_attributes.items():
            setattr(obj, k, kwargs.pop(k, v))

        return args, kwargs


class Base(metaclass=MetaBase):
    # Convertors
    def to_numpy(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_xarray(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_pandas(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_tfdataset(self, **kwargs):
        self._not_implemented()

    @abstractmethod
    def to_metview(self, **kwargs):
        self._not_implemented()

    # Change class
    def mutate(self):
        return self

    @classmethod
    def merge(cls, *args, **kwargs):
        return None

    # Used when plotting
    @abstractmethod
    def plot_map(self, backend):
        self._not_implemented()

    @abstractmethod
    def field_metadata(self):
        self._not_implemented()

    # I/O
    @abstractmethod
    def save(self, path):
        self._not_implemented()

    @abstractmethod
    def write(self, f):
        self._not_implemented()

    # Used by normalisers
    @abstractmethod
    def to_datetime(self):
        self._not_implemented()

    @abstractmethod
    def to_datetime_list(self):
        self._not_implemented()

    @abstractmethod
    def to_bounding_box(self):
        self._not_implemented()

    # For machine learning
    @abstractmethod
    def statistics(self):
        self._not_implemented()

    @abstractmethod
    def scaled(self, args, kwargs):
        self._not_implemented()

    @abstractmethod
    def sel(self, *args, **kwargs):
        self._not_implemented()

    @abstractmethod
    def order_by(self, *args, **kwargs):
        self._not_implemented()

    def to_pytorch_dataset(self, *args, **kwargs):
        self._not_implemented()

    def set_options(self, *args, **kwargs):
        if not hasattr(self, "_options"):
            self._options = {}
        for dic in args:
            self.set_options(**dic)
        if kwargs:
            self._options.update(**kwargs)

    def get_options(self):
        if not hasattr(self, "_options"):
            self._options = {}
        return self._options

    #
    def _not_implemented(self):
        import inspect

        func = inspect.stack()[1][3]
        module = self.__class__.__module__
        name = self.__class__.__name__

        extra = ""
        if hasattr(self, "path"):
            extra = f" on {self.path}"
        raise NotImplementedError(f"{module}.{name}.{func}(){extra}")



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import hashlib
import json
import logging
import os
import sqlite3
from abc import abstractmethod
from threading import local

import numpy as np

import climetlab as cml
from climetlab.utils import tqdm
from climetlab.utils.parts import Part

from . import (
    FILEPARTS_KEY_NAMES,
    MORE_KEY_NAMES,
    MORE_KEY_NAMES_WITH_UNDERSCORE,
    STATISTICS_KEY_NAMES,
    Database,
    FloatDBKey,
    IntDBKey,
    StrDBKey,
)

LOG = logging.getLogger(__name__)


def entryname_to_dbname(n):
    def add_mars(x):
        return "i_" + x

    def remove_first_underscore(x):
        assert x[0] == "_", x
        return x[1:]

    if n in FILEPARTS_KEY_NAMES:
        return remove_first_underscore(n)
    if n in MORE_KEY_NAMES_WITH_UNDERSCORE:
        return remove_first_underscore(n)
    if n in STATISTICS_KEY_NAMES:
        return n
    # if n in MORE_KEY_NAMES:
    #    return n
    return add_mars(n)


def dbname_to_entryname(n):
    if n.startswith("i_"):
        return n[2:]
    if "_" + n in FILEPARTS_KEY_NAMES:
        return "_" + n
    if "_" + n in MORE_KEY_NAMES_WITH_UNDERSCORE:
        return "_" + n
    if n in MORE_KEY_NAMES:
        return n
    return n


class EntriesLoader:
    table_name = "entries"

    def __init__(self, connection):
        self.connection = connection
        self.keys = self.read_from_table()

    @classmethod
    def guess_key_type(self, x, name=None):
        try:
            return {
                "TEXT": StrDBKey,
                "FLOAT": FloatDBKey,
                "INTEGER": IntDBKey,
                str: StrDBKey,
                float: FloatDBKey,
                np.float64: FloatDBKey,
                np.float32: FloatDBKey,
                int: IntDBKey,
                datetime.datetime: StrDBKey,
            }[x]
        except KeyError:
            raise ValueError(f"Unknown type '{x}' for key '{name}'.")

    def read_from_table(self):
        keys = {}
        cursor = self.connection.execute(f"PRAGMA table_info({self.table_name})")
        for x in cursor.fetchall():
            name, typ = x[1], x[2]
            klass = self.guess_key_type(typ, name)
            keys[name] = klass(name)

        if not keys:
            LOG.debug(f"Table {self.table_name} does not exist.")
        return keys

    def create_table_from_entry_if_needed(self, entry):
        keys = {}
        for k, v in entry.items():
            typ = type(v)
            klass = self.guess_key_type(typ, name=k)
            name = entryname_to_dbname(k)
            keys[name] = klass(name)

        assert keys, f"Cannot build from entry '{entry}'"
        LOG.debug(f"Created table {self} from entry {entry}.")
        columns_defs = ",".join([f"{v.name} {v.sql_type}" for k, v in keys.items()])
        statement = f"CREATE TABLE IF NOT EXISTS {self.table_name} ({columns_defs});"
        LOG.debug("%s", statement)
        self.connection.execute(statement)
        return keys

    def load_iterator(self, iterator):
        count = 0
        for entry in iterator:
            if count == 0:
                self.keys = self.create_table_from_entry_if_needed(entry)
                column_names = [k for k, v in self.keys.items()]
                statement = (
                    f"INSERT INTO {self.table_name} ("
                    + ",".join(column_names)
                    + ") VALUES("
                    + ",".join(["?"] * len(column_names))
                    + ");"
                )
                LOG.debug("%s", statement)

            values = [entry.get(dbname_to_entryname(k)) for k, v in self.keys.items()]
            self.connection.execute(statement, tuple(values))
            count += 1

        self.build_sql_indexes()

        return count

    def build_sql_indexes(self):
        indexed_columns = [k for k, v in self.keys.items() if k.startswith("i_")]
        indexed_columns += ["path"]

        pbar = tqdm(indexed_columns, desc="Building indexes")
        for n in pbar:
            pbar.set_description(f"Building index for {n}")
            self.connection.execute(
                f"CREATE INDEX IF NOT EXISTS {n}_index ON {self.table_name} ({n});"
            )

    def __str__(self):
        content = ",".join([k for k, v in self.keys.items()])
        return f"{self.__class__}({self.table_name},{content}"


class Connection(local):
    # Inheriting from threading.local allows one connection for each thread
    # __init__ is "called each time the local object is used in a separate thread".
    # https://github.com/python/cpython/blob/0346eddbe933b5f1f56151bdebf5bd49392bc275/Lib/_threading_local.py#L65
    def __init__(self, db_path):
        self._conn = sqlite3.connect(db_path)


class SqlFilter:
    def h(self, *args, **kwargs):
        m = hashlib.sha256()
        m.update(str(args).encode("utf-8"))
        m.update(str(kwargs).encode("utf-8"))
        m.update(str(self.__class__.__name__).encode("utf-8"))
        m.update(json.dumps(self.kwargs, sort_keys=True).encode("utf-8"))
        return m.hexdigest()

    def __str__(self):
        return f"{self.__class__.__name__}({self.kwargs})"

    @property
    def is_empty(self):
        return not self.kwargs

    @abstractmethod
    def filter_statement(self, db, *args, **kwargs):
        pass


class SqlSelection(SqlFilter):
    def __init__(self, kwargs: dict):
        self.kwargs = kwargs

    def filter_statement(self, db, *args, **kwargs):
        conditions = []
        for k, v in self.kwargs.items():
            if v is None or v == cml.ALL:
                continue

            name = entryname_to_dbname(k)
            dbkey = db.dbkeys[name]

            if not isinstance(v, (list, tuple)):
                v = [v]

            v = [dbkey.cast(x) for x in v]
            v = ["'" + str(x) + "'" for x in v]

            conditions.append(f"{name} IN ({', '.join(v)})")

        if not conditions:
            return ""
        return " WHERE " + " AND ".join(conditions)


class SqlOrder(SqlFilter):
    def __init__(self, kwargs):
        self.kwargs = kwargs

    def filter_statement(self, db, new_view):
        func_name = "userorder_" + new_view

        order_bys = []
        dict_of_dicts = {}

        for k, v in self.kwargs.items():
            name = entryname_to_dbname(k)

            dbkey = db.dbkeys[name]

            if v == "ascending" or v is None:
                order_bys.append(name + " ASC")
                continue
            if v == "descending" or v is None:
                order_bys.append(name + " DESC")
                continue
            if isinstance(v, (list, tuple)):
                v = [dbkey.cast(x) for x in v]
                dict_of_dicts[name] = dict(zip(v, range(len(v))))
                order_bys.append(func_name + "(" + "'" + name + "'" + ", " + name + ")")
                continue

            raise ValueError(f"{k},{v}, {type(v)}")

        if dict_of_dicts:

            def order_func(k, v):
                return dict_of_dicts[k][v]

            db.connection.create_function(func_name, 2, order_func, deterministic=True)

        if not order_bys:
            return ""
        return "ORDER BY " + ",".join(order_bys)


class VersionedDatabaseMixin:
    VERSION = 6

    @property
    def _version(self):
        cursor = self.connection.execute("PRAGMA user_version;")
        for res in cursor:
            version = res[0]
            return version if version else None
        assert False

    def _set_version(self):
        if self._version is None:
            self.connection.execute(f"PRAGMA user_version = {self.VERSION};")
            return
        self._check_version()

    def _check_version(self):
        version = self._version
        if version is None or version == self.VERSION:
            return
        raise Exception(
            (
                "Version mismatch: current version for database index"
                " is {self.VERSION} and the database already has version"
                f" {version}"
            )
        )


class SqlDatabase(Database, VersionedDatabaseMixin):
    EXTENSION = ".db"

    def __init__(
        self,
        db_path,
        filters=None,
    ):
        self._cache_column_names = {}

        self.db_path = db_path
        self._filters = filters or []
        self._view = None
        self._connection = None

        self.dbkeys = EntriesLoader(self.connection).keys

    @property
    def view(self):
        if self._view is None:
            self._view = "entries"
            for f in self._filters:
                self._apply_filter(f)
            LOG.debug("DB %s %s", self.db_path, self.view)
        return self._view

    @property
    def connection(self):
        if self._connection is None:
            self._connection = Connection(self.db_path)
        return self._connection._conn

    def _apply_filter(self, filter: SqlFilter):
        # This method updates self.view with the additional filter

        old_view = self._view
        new_view = old_view + "_" + filter.h(parent_view=old_view)

        filter_statement = filter.filter_statement(self, new_view)

        statement = (
            f"CREATE TEMP VIEW IF NOT EXISTS {new_view} AS SELECT * "
            + f"FROM {old_view} {filter_statement};"
        )

        LOG.debug("%s", statement)
        for i in self.connection.execute(statement):
            LOG.error(str(i))  # Output of .execute should be empty

        self._view = new_view

    def filter(self, filter: SqlFilter):
        return self.__class__(
            self.db_path,
            filters=self._filters + [filter],
        )

    def load(self, iterator):
        with self.connection as connection:
            loader = EntriesLoader(connection)
            count = loader.load_iterator(iterator)
            self.dbkeys = loader.keys

            assert count >= 1, "No entry found."
            LOG.info("Added %d entries", count)

        return count

    def lookup_parts(self, limit=None, offset=None, resolve_paths=True):
        """
        Look into the database and provide entries as Parts.
        limit: Returns only "limit" entries (used for paging).
        offset: Skip the first "offset" entries (used for paging).
        """

        _names = ["path", "offset", "length"]
        parts = []
        for path, offset, length in self._execute_select(_names, limit, offset):
            parts.append(Part(path, offset, length))
        if resolve_paths:
            parts = Part.resolve(parts, os.path.dirname(self.db_path))
        return parts

    def lookup_dicts(self, limit=None, offset=None, remove_none=True, with_parts=None):
        """
        Return entries dicts as they where inserted into the database.
        limit: Returns only "limit" entries (used for paging).
        offset: Skip the first "offset" entries (used for paging).
        """

        column_names = [k for k, v in self.dbkeys.items()]

        if with_parts is False:
            column_names = [k for k in column_names if k not in FILEPARTS_KEY_NAMES]

        entrynames = [dbname_to_entryname(k) for k in column_names]
        for tupl in self._execute_select(column_names, limit, offset):
            dic = {k: v for k, v in zip(entrynames, tupl)}

            if remove_none:
                dic = {k: v for k, v in dic.items() if v is not None}
            yield dic

    def _execute_select(self, column_names, limit=None, offset=None):
        names_str = ",".join([x for x in column_names]) if column_names else "*"
        limit_str = f" LIMIT {limit}" if limit is not None else ""
        offset_str = f" OFFSET {offset}" if offset is not None else ""

        statement = f"SELECT {names_str} FROM {self.view} {limit_str} {offset_str};"
        LOG.debug("%s", statement)

        for tupl in self.connection.execute(statement):
            yield tupl

    def count(self):
        statement = f"SELECT COUNT(*) FROM {self.view};"
        for result in self.connection.execute(statement):
            return result[0]
        assert False, statement  # Fail if result is empty.

    def duplicate_db(self, filename, **kwargs):
        new_db = SqlDatabase(db_path=filename)
        iterator = self.lookup_dicts()
        new_db.load(iterator)
        return new_db



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


class GribOutput:
    def __init__(self, filename, template=None, **kwargs):
        self.f = open(filename, "wb")
        self.template = template

    def write(self, values, metadata={}, template=None):
        if template is None:
            template = self.template

        handle = template.handle.clone()
        handle.set_values(values)
        for k, v in metadata.items():
            handle.set(k, v)
        handle.write(self.f)

    def close(self):
        self.f.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.close()


def new_grib_output(*args, **kwargs):
    return GribOutput(*args, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import warnings
from collections import defaultdict

from climetlab.readers.grib.index.sql import FieldsetInFilesWithSqlIndex
from climetlab.sources.indexed_directory import GenericDirectorySource
from climetlab.utils import progress_bar

LOG = logging.getLogger(__name__)

USE_REFERENCE = [
    "distinctLatitudes",
    "distinctLongitudes",
    "gridType",
    "NV",
    "Nx",
    "Ny",
    "paramId",
    "name",
    "shortName",
    "units",
    "dataType",
]
REMOVE = [
    "stepUnits",
    "endStep",
    "numberOfDirections",
    "numberOfFrequencies",
    "directionNumber",
    "frequencyNumber",
    "gridDefinitionDescription",
    "centre",
    "centreDescription",
    "edition",
    "subCentre",
    "stepType",
    "totalNumber",
    "cfName",
    "cfVarName",
    "missingValue",
    "typeOfLevel",
    "jScansPositively",
    "latitudeOfFirstGridPointInDegrees",
    "latitudeOfLastGridPointInDegrees",
    "longitudeOfFirstGridPointInDegrees",
    "longitudeOfLastGridPointInDegrees",
    "numberOfPoints",
    "iDirectionIncrementInDegrees",
    "jDirectionIncrementInDegrees",
    "iScansNegatively",
    "jPointsAreConsecutive",
]
METADATA_FUNCS = defaultdict(lambda x, item: item.owner.from_reference(x, item))

for k in REMOVE:
    METADATA_FUNCS[k] = lambda x, item: None

for k in USE_REFERENCE:
    METADATA_FUNCS[k] = lambda x, item: item.owner.from_reference(x, item)


def cast_or_none(typ, key):
    def f(x, item):
        value = item.item_metadata[key]
        if value is None:
            return None
        value = typ(value)
        return value

    return f


METADATA_FUNCS["param"] = lambda x, item: item.item_metadata["param"]
METADATA_FUNCS["shortName"] = lambda x, item: item.item_metadata["param"]
METADATA_FUNCS["number"] = cast_or_none(int, "number")
METADATA_FUNCS["level:float"] = cast_or_none(float, "levelist")
METADATA_FUNCS["level"] = cast_or_none(int, "levelist")
METADATA_FUNCS["dataDate"] = lambda x, item: int(item.item_metadata["date"])
METADATA_FUNCS["dataTime"] = lambda x, item: int(item.item_metadata["time"])


class NoLock:
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


class CacheDict(dict):
    def __init__(self, field):
        self.field = field

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            self[key] = self.field[key]
            return self[key]

    def __str__(self):
        return f"{self.field} (cached)"


class VirtualField:  # Should inherit from GribField
    _real_item = None

    def __init__(self, i, owner):
        self.i = i
        self.owner = owner
        self.item_metadata = owner.get_metadata(i)

    @property
    def DEBUG(self):
        return self.owner.DEBUG

    @property
    def real_item(self):
        if self._real_item is None:
            self._real_item = self.owner.get_real_item(self.i)
        return self._real_item

    def write(self, f):
        return self.real_item.write(f)

    def metadata(self, key):
        return self.owner._get_metadata_for_item(key, self)

    @property
    def values(self):
        return self.real_item.to_numpy()

    def to_numpy(self):
        return self.real_item.to_numpy()

    @property
    def shape(self):
        return self.owner.reference[0].shape

    def __str__(self):
        return "Virt" + str(self.real_item)


class VirtualFieldsetInFilesWithSqlIndex(FieldsetInFilesWithSqlIndex):
    # DEBUG = True
    DEBUG = False

    def __init__(self, *args, **kwargs):
        self._reference = None
        self.pbar = None
        super().__init__(*args, **kwargs)

    def xarray_open_dataset_kwargs(self):
        return dict(
            cache=False,  # Set to False to prevent loading the whole dataset
            # chunks={ },
            lock=NoLock(),
        )

    def __getitem__(self, n):
        if n >= len(self):
            raise IndexError
        if self.pbar:
            self.pbar.update(1)

        item = VirtualField(n, owner=self)

        self.check_same_metadata_as_reference("md5_grid_section", item)
        self.check_same_metadata_as_reference("param", item)

        return item

    def get_real_item(self, n):
        return super().__getitem__(n)

    def _get_metadata_for_item(self, key, item):
        func = METADATA_FUNCS[key]

        try:
            value = func(key, item)
        except Exception:
            # except Exception as e:
            # LOG.exception(f"DEBUG MESSAGE: Exception reading {key}:{str(e)}")
            return None

        if key not in METADATA_FUNCS:
            warnings.warn(f"Reading from reference (not expected): {k}={func(k)}")

        if item.DEBUG:
            if key not in REMOVE:
                self._check_with_real_item(key, item, value)
            else:
                assert value is None, (key, value)

        return value

    def to_xarray(self, *args, **kwargs):
        self.pbar = progress_bar(desc="to xarray()", total=len(self))
        ds = super().to_xarray(*args, **kwargs)
        self.pbar = None
        return ds

    @property
    def reference(self):
        if self._reference is None:
            reference = self.get_real_item(0)
            metadata = self.get_metadata(0)
            self._reference = (CacheDict(reference), metadata)
        return self._reference

    def from_reference(self, key, item):
        value = self.reference[0][key]
        return value

    def _check_with_real_item(self, key, item, value):
        assert self.DEBUG is True
        real = item.real_item[key]
        if type(real) != type(value) or str(real) != str(value):
            msg = f"key={key}: providing {value} (type {type(value)}) instead of {real} (type {type(real)})"
            warnings.warn(msg)
            exit(-1)
        return True

    def check_same_metadata_as_reference(self, key, item):
        r = self.reference[1][key]
        i = item.item_metadata[key]
        if r != i:
            raise Exception(
                (
                    "The virtual-directory source does not support using multiple parameter in the same source yet."
                    f"Error for field={item.i} and {key=}: the reference field metadata is {r=}"
                    f" but the metatada for the item is {i=}"
                )
            )


class VirtualDirectorySource(GenericDirectorySource):
    INDEX_CLASS = VirtualFieldsetInFilesWithSqlIndex

    def __init__(self, *args, **kwargs):
        print(
            'Source "virtual-directory" is deprecated. Use "indexed-directory" instead.'
        )
        LOG.warn(
            'Source "virtual-directory" is deprecated. Use "indexed-directory" instead.'
        )
        super().__init__(*args, **kwargs)


source = VirtualDirectorySource



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import json
import logging
import os

from climetlab.indexing.database.json import JsonStdoutDatabase
from climetlab.readers.grib.parsing import (
    GribIndexingDirectoryParserIterator,
    _index_url,
)
from climetlab.utils.humanize import plural, seconds

from .tools import parse_args

LOG = logging.getLogger(__name__)


class GribCmd:
    @parse_args(
        urls=(
            None,
            dict(metavar="URLS", type=str, nargs="+", help="List of urls to index."),
        ),
        baseurl=dict(
            metavar="BASEURL",
            type=str,
            help="Base url to use as a prefix to happen on each URLS to build actual urls.",
        ),
    )
    def do_index_urls(self, args):
        """Create json index files for remote Grib urls.
        If the option --baseurl is provided, the given url are relative to the BASEURL.
        This allows creating an index for multiple gribs."""

        if args.baseurl:

            def resolve_url(u):
                if u.startswith(args.baseurl):
                    # already absolute url
                    return u
                return f"{args.baseurl}/{u}"

        else:

            def resolve_url(url):
                return url

        stdout = JsonStdoutDatabase()
        for u in args.urls:
            url = resolve_url(u)
            for e in _index_url(url):
                e["_path"] = u
                stdout.load([e])

    @parse_args(
        url=(None, dict(metavar="URL", type=str, help="url to index")),
        # output=dict(type=str, help="Output filename"),
    )
    def do_index_url(self, args):
        """Create json index files for remote Grib url."""
        stdout = JsonStdoutDatabase()
        for e in _index_url(args.url):
            assert "_path" not in e
            stdout.load([e])

    @parse_args(
        directory=(None, dict(help="Directory containing the GRIB files to index.")),
        format=dict(
            default="sql",
            metavar="FORMAT",
            type=str,
            help="sql or json or stdout.",
        ),
        force=dict(action="store_true", help="overwrite existing index."),
        # pattern=dict(help="Files to index (patterns).", nargs="*"),
        no_follow_links=dict(action="store_true", help="Do not follow symlinks."),
        relative_paths=dict(
            action="store_true",
            help=(
                "Use relative paths. Default is to use relative paths, "
                "except when a custom location is provided for the index location. "
                "(when the argument --output is used, default for --relative-path is False)"
            ),
        ),
        output=(
            "--output",
            dict(
                help="Custom location of the database file, will write absolute filenames in the database."
            ),
        ),
    )
    def do_index_directory(self, args):
        """Index a directory containing GRIB files."""
        directory = args.directory
        db_path = args.output
        force = args.force
        force_relative_paths_on = args.relative_paths
        db_format = args.format

        followlinks = True
        if args.no_follow_links:
            followlinks = False

        assert os.path.isdir(directory), directory

        from climetlab.sources.indexed_directory import IndexedDirectorySource

        if force_relative_paths_on:
            relative_paths = True
        else:
            if db_path is None:
                relative_paths = True
            else:
                LOG.warning(
                    (
                        "Non-default location for the index file (--output),"
                        " using absolute path in the index (i.e. setting option --relative-paths to False)."
                    )
                )
                relative_paths = False

        def default_db_path():
            path = dict(
                json=os.path.join(directory, IndexedDirectorySource.DEFAULT_JSON_FILE),
                sql=os.path.join(directory, IndexedDirectorySource.DEFAULT_DB_FILE),
                stdout=None,
            )[db_format]
            return path

        if db_path is None:
            db_path = default_db_path()

        def check_overwrite(filename, force):
            if not os.path.exists(filename):
                return
            if force:
                print(f"File {filename} already exists, overwriting it.")
                os.unlink(filename)
                return
            raise Exception(
                f"ERROR: File {filename} already exists (use --force to overwrite)."
            )

        if db_path is not None:
            check_overwrite(db_path, force)

        ignore = []
        ignore.append("climetlab*.db")
        ignore.append("climetlab*.json")
        # Do not add anything more to this ignore list, implement a pattern filter if needed.
        ignore.append("*.idx")
        if db_path is not None:
            ignore.append(db_path)

        _index_directory(
            directory,
            db_path=db_path,
            relative_paths=relative_paths,
            followlinks=followlinks,
            ignore=ignore,
            db_format=db_format,
            with_statistics=True,
        )

    @parse_args(
        filename=(None, dict(help="Database filename.")),
    )
    def do_dump_index(self, args):
        db_path = args.filename
        from climetlab.sources.indexed_directory import IndexedDirectorySource

        if os.path.isdir(db_path):
            db_path = os.path.join(db_path, IndexedDirectorySource.DEFAULT_DB_FILE)

        from climetlab.indexing.database.sql import SqlDatabase

        db = SqlDatabase(db_path=db_path)
        for d in db.lookup_dicts():
            print(json.dumps(d))


def _index_directory(
    directory,
    db_path,
    relative_paths,
    followlinks,
    ignore,
    db_format,
    with_statistics,
):
    from climetlab.indexing.database.json import JsonFileDatabase, JsonStdoutDatabase
    from climetlab.indexing.database.sql import SqlDatabase

    db = dict(
        json=JsonFileDatabase,
        sql=SqlDatabase,
        stdout=JsonStdoutDatabase,
    )[
        db_format
    ](db_path)
    iterator = GribIndexingDirectoryParserIterator(
        directory,
        ignore=ignore,
        relative_paths=relative_paths,
        followlinks=followlinks,
        with_statistics=with_statistics,
    )
    start = datetime.datetime.now()
    count = db.load(iterator)
    end = datetime.datetime.now()
    print(f"Indexed {plural(count,'field')} in {seconds(end - start)}.")



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

import climetlab as cml

from .tools import parse_args

LOG = logging.getLogger(__name__)


class AvailabilityCmd:
    @parse_args(
        source=(
            None,
            dict(
                type=str,
                help="File or directory for describing a dataset or source of data with GRIB data.",
            ),
        ),
        stdout=dict(action="store_true", help="Output to stdout (no file)."),
    )
    def do_availability(self, args):
        """Create json availability file."""

        self.avail = None

        path = args.source

        if not os.path.exists(path):
            print(f"{path} does not exists.")

        if os.path.isdir(path):
            source = availability_of_directory(path)
        else:
            source = availability_of_file(path)

        if args.stdout:
            print(source.availability.tree())
            return

        self.write(source.availability, source.availability_path)

    def write(self, avail, output):
        if os.path.exists(output):
            i = 1
            while os.path.exists(f"{output}.{i}"):
                i += 1
            output = f"{output}.{i}"
            print(f"File already exists, writing to {output}")

        avail.to_pickle(output)


def availability_of_directory(dirpath):
    db_path = os.path.join(dirpath, "climetlab-2.db")
    if not os.path.exists(db_path):
        print(f"ERROR: this directory is not indexed yet, cannot find {db_path}.")

    source = cml.load_source("indexed-directory", dirpath)
    return source


def availability_of_file(path):
    # if magic = "SQLite"...
    return cml.load_source("file", path)
